from typing import List, Type

from sqlalchemy.inspection import inspect
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.orm.decl_api import DeclarativeMeta
from sqlalchemy.orm.relationships import RelationshipProperty
from sqlalchemy.sql.expression import BinaryExpression, ClauseElement, ColumnClause

from odata_query import ast, exceptions as ex, utils, visitor

from . import common


class AstToSqlAlchemyOrmVisitor(common._CommonVisitors, visitor.NodeVisitor):
    """
    :class:`NodeVisitor` that transforms an :term:`AST` into a SQLAlchemy query
    filter clause using ORM features.

    Args:
        root_model: The root model of the query.
    """

    def __init__(self, root_model: Type[DeclarativeMeta]):
        self.root_model = root_model
        self.join_relationships: List[InstrumentedAttribute] = []

    def visit_Identifier(self, node: ast.Identifier) -> ColumnClause:
        ":meta private:"
        try:
            return getattr(self.root_model, node.name)
        except AttributeError:
            raise ex.InvalidFieldException(node.name)

    def visit_Attribute(self, node: ast.Attribute) -> ColumnClause:
        ":meta private:"
        rel_attr = self.visit(node.owner)
        # Owner is an InstrumentedAttribute, hopefully of a relationship.
        # But we need the model pointed to by the relationship.
        prop_inspect = inspect(rel_attr).property
        if not isinstance(prop_inspect, RelationshipProperty):
            # TODO: new exception:
            raise ValueError(f"Not a relationship: {node.owner}")
        self.join_relationships.append(rel_attr)

        # We'd like to reference the column on the related class:
        owner_cls = prop_inspect.entity.class_
        try:
            return getattr(owner_cls, node.attr)
        except AttributeError:
            raise ex.InvalidFieldException(node.attr)

    def visit_Compare(self, node: ast.Compare) -> BinaryExpression:
        ":meta private:"
        left = self.visit(node.left)
        right = self.visit(node.right)
        op = self.visit(node.comparator)

        # If a node is a `relationship` representing a single foreign key,
        # the client meant to compare the foreign key, not the related object.
        # E.g. In "blogpost/author eq 1", left should be "blogpost/author_id"
        left = self._maybe_sub_relationship_with_foreign_key(left)
        right = self._maybe_sub_relationship_with_foreign_key(right)

        return op(left, right)

    def visit_CollectionLambda(self, node: ast.CollectionLambda) -> ClauseElement:
        ":meta private:"
        owner_prop = self.visit(node.owner)
        collection_model = inspect(owner_prop).property.entity.class_

        if node.lambda_:
            # For the lambda, we want to strip the identifier off, because
            # we will execute this as a subquery in the wanted model's context.
            subq_ast = utils.expression_relative_to_identifier(
                node.lambda_.identifier, node.lambda_.expression
            )
            subq_transformer = self.__class__(collection_model)
            subquery_filter = subq_transformer.visit(subq_ast)
        else:
            subquery_filter = None

        if isinstance(node.operator, ast.Any):
            return owner_prop.any(subquery_filter)
        else:
            # For an ALL query, invert both the filter and the EXISTS:
            if node.lambda_:
                subquery_filter = ~subquery_filter
            return ~owner_prop.any(subquery_filter)

    def _maybe_sub_relationship_with_foreign_key(
        self, elem: ClauseElement
    ) -> ClauseElement:
        """
        If the given ClauseElement is a `relationship` with a single ForeignKey,
        replace it with the `ForeignKey` itself.

        :meta private:
        """
        try:
            prop_inspect = inspect(elem).property
            if isinstance(prop_inspect, RelationshipProperty):
                foreign_key = prop_inspect._calculated_foreign_keys
                if len(foreign_key) == 1:
                    return next(iter(foreign_key))
        except Exception:
            pass

        return elem



from typing import List

from sqlalchemy.orm.query import Query
from sqlalchemy.sql.expression import ClauseElement, Select

from odata_query.grammar import ODataLexer, ODataParser  # type: ignore

from .core import AstToSqlAlchemyCoreVisitor
from .orm import AstToSqlAlchemyOrmVisitor


def _get_joined_attrs(query: Select) -> List[str]:
    # use _legacy_setup_joins for legacy Query objects
    setup_joins = (
        getattr(query, "_legacy_setup_joins", query._setup_joins) or query._setup_joins
    )
    return [str(join[0]) for join in setup_joins]


def apply_odata_query(query: ClauseElement, odata_query: str) -> ClauseElement:
    """
    Shorthand for applying an OData query to a SQLAlchemy query.

    Args:
        query: SQLAlchemy query to apply the OData query to.
        odata_query: OData query string.
    Returns:
        ClauseElement: The modified query
    """
    lexer = ODataLexer()
    parser = ODataParser()

    clause_elem: ClauseElement
    if isinstance(query, Query):
        # For now, we keep supporting the 1.x style of queries unofficially.
        # GITHUB-34
        clause_elem = query.__clause_element__()
    else:
        clause_elem = query

    model = clause_elem.columns_clause_froms[0].entity_namespace

    ast = parser.parse(lexer.tokenize(odata_query))
    transformer = AstToSqlAlchemyOrmVisitor(model)
    where_clause = transformer.visit(ast)

    existing_joins = _get_joined_attrs(query)
    for required_join in transformer.join_relationships:
        if (
            str(required_join) not in existing_joins
            and str(required_join.key) not in existing_joins
        ):
            query = query.join(required_join)

    return query.filter(where_clause)


def apply_odata_core(query: ClauseElement, odata_query: str) -> ClauseElement:
    """
    Shorthand for applying an OData query to a SQLAlchemy core.

    Args:
        query: SQLAlchemy query to apply the OData query to.
        odata_query: OData query string.
    Returns:
        ClauseElement: The modified query
    """
    lexer = ODataLexer()
    parser = ODataParser()
    table = query.columns_clause_froms[0]

    ast = parser.parse(lexer.tokenize(odata_query))
    transformer = AstToSqlAlchemyCoreVisitor(table)
    where_clause = transformer.visit(ast)
    return query.filter(where_clause)



from odata_query import ast, exceptions, typing

from .base import AstToSqlVisitor


class AstToSqliteSqlVisitor(AstToSqlVisitor):
    """
    :class:`NodeVisitor` that transforms an :term:`AST` into a SQLite SQL
    ``WHERE`` clause.

    Args:
        table_alias: Optional alias for the root table.
    """

    def visit_Boolean(self, node: ast.Boolean) -> str:
        """:meta private:"""
        if node.py_val:
            return "1"
        return "0"

    def visit_Date(self, node: ast.Date) -> str:
        """:meta private:"""
        return f"DATE('{node.val}')"

    def visit_DateTime(self, node: ast.DateTime) -> str:
        """:meta private:"""
        return f"DATETIME('{node.val}')"

    def sqlfunc_indexof(self, *args: ast._Node) -> str:
        """:meta private:"""
        args_sql = [self.visit(arg) for arg in args]
        inferred_type = [typing.infer_type(arg) for arg in args]

        # If any of the inputs is a string, assume str-indexof:
        if any(typ is ast.String for typ in inferred_type) or all(
            typ is None for typ in inferred_type
        ):
            return f"INSTR({args_sql[0]}, {args_sql[1]}) - 1"

        # If any of the inputs is a list, assume list-indexof
        # which isn't easily doable at the moment:
        if any(typ is ast.List for typ in inferred_type):
            raise exceptions.UnsupportedFunctionException("indexof<List>")

        raise exceptions.ArgumentTypeException("indexof")

    def sqlfunc_length(self, arg: ast._Node) -> str:
        """:meta private:"""
        arg_sql = self.visit(arg)
        return f"LENGTH({arg_sql})"

    def sqlfunc_substring(self, *args: ast._Node) -> str:
        """:meta private:"""
        args_sql = [self.visit(arg) for arg in args]
        inferred_type = typing.infer_type(args[0])

        # If the first input is a string or default, assume str-substr:
        if inferred_type is ast.String or inferred_type is None:
            if len(args) == 2:
                return f"SUBSTR({args_sql[0]}, {args_sql[1]} + 1)"
            if len(args) == 3:
                return f"SUBSTR({args_sql[0]}, {args_sql[1]} + 1, {args_sql[2]})"

        # If the first input is a list, assume list-substr:
        if inferred_type is ast.List:
            raise exceptions.UnsupportedFunctionException("substring<List>")

        raise exceptions.ArgumentTypeException("substring")

    def sqlfunc_year(self, arg: ast._Node) -> str:
        """:meta private:"""
        arg_sql = self.visit(arg)
        return f"CAST(STRFTIME('%Y', {arg_sql}) AS INTEGER)"

    def sqlfunc_month(self, arg: ast._Node) -> str:
        """:meta private:"""
        arg_sql = self.visit(arg)
        return f"CAST(STRFTIME('%m', {arg_sql}) AS INTEGER)"

    def sqlfunc_day(self, arg: ast._Node) -> str:
        """:meta private:"""
        arg_sql = self.visit(arg)
        return f"CAST(STRFTIME('%d', {arg_sql}) AS INTEGER)"

    def sqlfunc_hour(self, arg: ast._Node) -> str:
        """:meta private:"""
        arg_sql = self.visit(arg)
        return f"CAST(STRFTIME('%H', {arg_sql}) AS INTEGER)"

    def sqlfunc_minute(self, arg: ast._Node) -> str:
        """:meta private:"""
        arg_sql = self.visit(arg)
        return f"CAST(STRFTIME('%M', {arg_sql}) AS INTEGER)"

    def sqlfunc_date(self, arg: ast._Node) -> str:
        """:meta private:"""
        arg_sql = self.visit(arg)
        return f"DATE({arg_sql})"

    def sqlfunc_now(self) -> str:
        """:meta private:"""
        return "DATETIME('now')"

    def sqlfunc_round(self, arg: ast._Node) -> str:
        """:meta private:"""
        arg_sql = self.visit(arg)
        return f"TRUNC({arg_sql} + 0.5)"

    def sqlfunc_floor(self, arg: ast._Node) -> str:
        """:meta private:"""
        arg_sql = self.visit(arg)
        return f"FLOOR({arg_sql})"

    def sqlfunc_ceiling(self, arg: ast._Node) -> str:
        """:meta private:"""
        arg_sql = self.visit(arg)
        return f"CEILING({arg_sql})"



from typing import Type

import sqlalchemy as sa
from sqlalchemy.sql.expression import BinaryExpression, ClauseElement, ColumnClause

from odata_query import ast, exceptions as ex, visitor

from . import common


class AstToSqlAlchemyCoreVisitor(common._CommonVisitors, visitor.NodeVisitor):
    """
    :class:`NodeVisitor` that transforms an :term:`AST` into a SQLAlchemy where
    clause using Core features.

    Args:
        table: A SQLalchemy table
    """

    def __init__(self, table: Type[sa.Table]):
        self.table = table

    def visit_Identifier(self, node: ast.Identifier) -> ColumnClause:
        """:meta private:"""
        try:
            return self.table.c[node.name]
        except KeyError:
            raise ex.InvalidFieldException(node.name)

    def visit_Attribute(self, node: ast.Attribute) -> ColumnClause:
        """:meta private:"""
        raise NotImplementedError(
            "Relationship traversal is not yet implemented for the SQLAlchemy Core visitor."
        )

    def visit_Compare(self, node: ast.Compare) -> BinaryExpression:
        """:meta private:"""
        left = self.visit(node.left)
        right = self.visit(node.right)
        op = self.visit(node.comparator)
        return op(left, right)

    def visit_CollectionLambda(self, node: ast.CollectionLambda) -> ClauseElement:
        """:meta private:"""
        raise NotImplementedError(
            "Collection lambda is not yet implemented for the SQLAlchemy Core visitor."
        )



import logging
from typing import Optional

from odata_query import ast, exceptions, typing, visitor

log = logging.getLogger(__name__)


class AstToSqlVisitor(visitor.NodeVisitor):
    """
    :class:`NodeVisitor` that transforms an :term:`AST` into a SQL ``WHERE``
    clause. Based on SQL-99 as described here: https://crate.io/docs/sql-99/en/latest/

    Args:
        table_alias: Optional alias for the root table.
    """

    def __init__(self, table_alias: Optional[str] = None):
        super().__init__()
        self.table_alias = table_alias

    def visit_Identifier(self, node: ast.Identifier) -> str:
        ":meta private:"
        # Double quotes for column names acc SQL Standard
        sql_id = f'"{node.name}"'

        if self.table_alias:
            sql_id = f'"{self.table_alias}".' + sql_id

        return sql_id

    def visit_Null(self, node: ast.Null) -> str:
        ":meta private:"
        return "NULL"

    def visit_Integer(self, node: ast.Integer) -> str:
        ":meta private:"
        return node.val

    def visit_Float(self, node: ast.Float) -> str:
        ":meta private:"
        return node.val

    def visit_Boolean(self, node: ast.Boolean) -> str:
        ":meta private:"
        return node.val.upper()

    def visit_String(self, node: ast.String) -> str:
        ":meta private:"
        # Replace single quotes with double single-quotes acc SQL standard:
        val = node.val.replace("'", "''")
        # Wrap in single quotes for string constants acc SQL Standard
        return f"'{val}'"

    def visit_Date(self, node: ast.Date) -> str:
        ":meta private:"
        # Single quotes for date constants acc SQL Standard
        return f"DATE '{node.val}'"

    def visit_DateTime(self, node: ast.DateTime) -> str:
        ":meta private:"
        sql_ts = node.val.replace("T", " ")
        # Single quotes for datetime constants acc SQL Standard
        return f"TIMESTAMP '{sql_ts}'"

    def visit_Duration(self, node: ast.Duration) -> str:
        ":meta private:"
        sign, days, hours, minutes, seconds = node.unpack()

        sign = sign or ""
        intervals = []
        if days:
            intervals.append(f"INTERVAL '{days}' DAY")
        if hours:
            intervals.append(f"INTERVAL '{hours}' HOUR")
        if minutes:
            intervals.append(f"INTERVAL '{minutes}' MINUTE")
        if seconds:
            intervals.append(f"INTERVAL '{seconds}' SECOND")

        if len(intervals) == 0:
            # Shouldn't occur but whatever
            return ""
        if len(intervals) == 1:
            return f"{sign}{intervals[0]}"
        if len(intervals) > 1:
            interval = " + ".join(intervals)
            return f"{sign}({interval})"

        # Make Quality checks happy:
        raise Exception("This code is never reachable...")

    def visit_GUID(self, node: ast.GUID) -> str:
        ":meta private:"
        return f"'{node.val}'"

    def visit_List(self, node: ast.List) -> str:
        ":meta private:"
        options = ", ".join(self.visit(n) for n in node.val)
        return f"({options})"

    def visit_Add(self, node: ast.Add) -> str:
        ":meta private:"
        return "+"

    def visit_Sub(self, node: ast.Sub) -> str:
        ":meta private:"
        return "-"

    def visit_Mult(self, node: ast.Mult) -> str:
        ":meta private:"
        return "*"

    def visit_Div(self, node: ast.Div) -> str:
        ":meta private:"
        return "/"

    def visit_Mod(self, node: ast.Mod) -> str:
        ":meta private:"
        return "%"

    def visit_BinOp(self, node: ast.BinOp) -> str:
        ":meta private:"
        left = self.visit(node.left)
        right = self.visit(node.right)
        op = self.visit(node.op)

        return f"{left} {op} {right}"

    def visit_Eq(self, node: ast.Eq) -> str:
        ":meta private:"
        return "="

    def visit_NotEq(self, node: ast.NotEq) -> str:
        ":meta private:"
        return "!="

    def visit_Lt(self, node: ast.Lt) -> str:
        ":meta private:"
        return "<"

    def visit_LtE(self, node: ast.LtE) -> str:
        ":meta private:"
        return "<="

    def visit_Gt(self, node: ast.Gt) -> str:
        ":meta private:"
        return ">"

    def visit_GtE(self, node: ast.GtE) -> str:
        ":meta private:"
        return ">="

    def visit_In(self, node: ast.In) -> str:
        ":meta private:"
        return "IN"

    def visit_Compare(self, node: ast.Compare) -> str:
        ":meta private:"
        left = self.visit(node.left)
        right = self.visit(node.right)
        comparator = self.visit(node.comparator)

        # In case of a subexpression, wrap it in parentheses
        if isinstance(node.left, (ast.BoolOp, ast.Compare)):
            left = f"({left})"
        if isinstance(node.right, (ast.BoolOp, ast.Compare)):
            right = f"({right})"

        #  'eq/ne null' should become 'IS (NOT) NULL' instead of '(!)= NULL'
        if isinstance(node.right, ast.Null):
            if isinstance(node.comparator, ast.Eq):
                comparator = "IS"
            elif isinstance(node.comparator, ast.NotEq):
                comparator = "IS NOT"

        return f"{left} {comparator} {right}"

    def visit_And(self, node: ast.And) -> str:
        ":meta private:"
        return "AND"

    def visit_Or(self, node: ast.Or) -> str:
        ":meta private:"
        return "OR"

    def visit_BoolOp(self, node: ast.BoolOp) -> str:
        ":meta private:"
        left = self.visit(node.left)
        op = self.visit(node.op)
        right = self.visit(node.right)

        # In case of a subexpression, wrap it in parentheses
        # UNLESS it has the same operator as the current BoolOp, e.g.:
        # x AND y AND z
        if isinstance(node.left, ast.BoolOp) and node.left.op != node.op:
            left = f"({left})"
        if isinstance(node.right, ast.BoolOp) and node.right.op != node.op:
            right = f"({right})"

        return f"{left} {op} {right}"

    def visit_Not(self, node: ast.Not) -> str:
        ":meta private:"
        return "NOT"

    def visit_UnaryOp(self, node: ast.UnaryOp) -> str:
        ":meta private:"
        op = self.visit(node.op)
        operand = self.visit(node.operand)

        # In case of a subexpression, wrap it in parentheses
        if isinstance(node.operand, ast.BoolOp):
            operand = f"({operand})"

        return f"{op} {operand}"

    def visit_Call(self, node: ast.Call) -> str:
        ":meta private:"
        try:
            # Grammar has already validated that the function is valid OData,
            # but that doesn't guarantee we can represent it in SQL:
            sql_gen = getattr(self, "sqlfunc_" + node.func.name.lower())
        except AttributeError:
            raise exceptions.UnsupportedFunctionException(node.func.name)

        return sql_gen(*node.args)

    def sqlfunc_concat(self, *args: ast._Node) -> str:
        ":meta private:"
        args_sql = [self.visit(arg) for arg in args]
        return f"{args_sql[0]} || {args_sql[1]}"

    def _to_pattern(self, arg: ast._Node, prefix: str = "", suffix: str = "") -> str:
        """
        Transform a node into a pattern usable in `LIKE` clauses.
        :meta private:
        """
        if isinstance(arg, (ast.Identifier, ast.Call)):
            res = self.visit(arg)
            if prefix:
                res = f"'{prefix}' || " + res
            if suffix:
                res = res + f" || '{suffix}'"
        else:
            res = str(arg.val).replace("%", "%%").replace("_", "__")  # type: ignore
            res = "'" + prefix + res + suffix + "'"
        return res

    def sqlfunc_contains(self, *args: ast._Node) -> str:
        ":meta private:"
        args_sql = [self.visit(arg) for arg in args]
        inferred_type = [typing.infer_type(arg) for arg in args]

        # If any of the inputs is a string or default, assume str-contains:
        if any(typ is ast.String for typ in inferred_type) or all(
            typ is None for typ in inferred_type
        ):
            pattern = self._to_pattern(args[1], prefix="%", suffix="%")
            return f"{args_sql[0]} LIKE {pattern}"

        # If any of the inputs is a list, assume list-contains:
        if any(typ is ast.List for typ in inferred_type):
            raise exceptions.UnsupportedFunctionException("contains<List>")

        raise exceptions.ArgumentTypeException("contains")

    def sqlfunc_endswith(self, *args: ast._Node) -> str:
        ":meta private:"
        args_sql = [self.visit(arg) for arg in args]
        inferred_type = [typing.infer_type(arg) for arg in args]

        # If any of the inputs is a string or default, assume str-endswith:
        if any(typ is ast.String for typ in inferred_type) or all(
            typ is None for typ in inferred_type
        ):
            pattern = self._to_pattern(args[1], prefix="%")
            return f"{args_sql[0]} LIKE {pattern}"

        # If any of the inputs is a list, assume list-endswith
        # which isn't easily doable at the moment:
        if any(typ is ast.List for typ in inferred_type):
            raise exceptions.UnsupportedFunctionException("endswith<List>")

        raise exceptions.ArgumentTypeException("endswith")

    def sqlfunc_indexof(self, *args: ast._Node) -> str:
        ":meta private:"
        args_sql = [self.visit(arg) for arg in args]
        inferred_type = [typing.infer_type(arg) for arg in args]

        # If any of the inputs is a string, assume str-indexof:
        if any(typ is ast.String for typ in inferred_type) or all(
            typ is None for typ in inferred_type
        ):
            return f"POSITION({args_sql[1]} IN {args_sql[0]}) - 1"

        # If any of the inputs is a list, assume list-indexof
        # which isn't easily doable at the moment:
        if any(typ is ast.List for typ in inferred_type):
            raise exceptions.UnsupportedFunctionException("indexof<List>")

        raise exceptions.ArgumentTypeException("indexof")

    def sqlfunc_length(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        inferred_type = typing.infer_type(arg)

        # If the input is a string or default, assume str-length:
        if inferred_type is ast.String or inferred_type is None:
            return f"CHAR_LENGTH({arg_sql})"

        # If the input is a list, assume list-length:
        if inferred_type is ast.List:
            return f"CARDINALITY({arg_sql})"

        raise exceptions.ArgumentTypeException("length")

    def sqlfunc_startswith(self, *args: ast._Node) -> str:
        ":meta private:"
        args_sql = [self.visit(arg) for arg in args]
        inferred_type = [typing.infer_type(arg) for arg in args]

        # If any of the inputs is a string or default, assume str-startswith:
        if any(typ is ast.String for typ in inferred_type) or all(
            typ is None for typ in inferred_type
        ):
            pattern = self._to_pattern(args[1], suffix="%")
            return f"{args_sql[0]} LIKE {pattern}"

        # If any of the inputs is a list, assume list-startswith
        # which isn't easily doable at the moment:
        if any(typ is ast.List for typ in inferred_type):
            raise exceptions.UnsupportedFunctionException("startswith<List>")

        raise exceptions.ArgumentTypeException("startswith")

    def sqlfunc_substring(self, *args: ast._Node) -> str:
        ":meta private:"
        args_sql = [self.visit(arg) for arg in args]
        inferred_type = typing.infer_type(args[0])

        # If the first input is a string or default, assume str-substr:
        if inferred_type is ast.String or inferred_type is None:
            if len(args) == 2:
                return f"SUBSTRING({args_sql[0]} FROM {args_sql[1]} + 1)"
            if len(args) == 3:
                return (
                    f"SUBSTRING({args_sql[0]} FROM {args_sql[1]} + 1 FOR {args_sql[2]})"
                )

        # If the first input is a list, assume list-substr:
        if inferred_type is ast.List:
            raise exceptions.UnsupportedFunctionException("substring<List>")

        raise exceptions.ArgumentTypeException("substring")

    def sqlfunc_tolower(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"LOWER({arg_sql})"

    def sqlfunc_toupper(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"UPPER({arg_sql})"

    def sqlfunc_trim(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"TRIM({arg_sql})"

    def sqlfunc_year(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"EXTRACT (YEAR FROM {arg_sql})"

    def sqlfunc_month(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"EXTRACT (MONTH FROM {arg_sql})"

    def sqlfunc_day(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"EXTRACT (DAY FROM {arg_sql})"

    def sqlfunc_hour(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"EXTRACT (HOUR FROM {arg_sql})"

    def sqlfunc_minute(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"EXTRACT (MINUTE FROM {arg_sql})"

    def sqlfunc_date(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"CAST ({arg_sql} AS DATE)"

    def sqlfunc_now(self) -> str:
        ":meta private:"
        return "CURRENT_TIMESTAMP"

    def sqlfunc_round(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"CAST ({arg_sql} + 0.5 AS INTEGER)"

    def sqlfunc_floor(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"""CASE {arg_sql}
    WHEN > 0 CAST ({arg_sql} AS INTEGER)
    WHEN < 0 CAST (0 - (ABS({arg_sql}) + 0.5) AS INTEGER))
    ELSE {arg_sql}
END"""

    def sqlfunc_ceiling(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"""CASE {arg_sql} - CAST ({arg_sql} AS INTEGER)
    WHEN > 0 {arg_sql}+1
    WHEN < 0 {arg_sql}-1
    ELSE {arg_sql}
END"""

    def sqlfunc_hassubset(self, *args: ast._Node) -> str:
        ":meta private:"
        raise exceptions.UnsupportedFunctionException("hassubset")



from sqlalchemy.sql.functions import GenericFunction
from sqlalchemy.types import Integer, String


class strpos(GenericFunction):
    type = Integer
    package = "odata"
    inherit_cache = True


class substr(GenericFunction):
    type = String
    package = "odata"
    inherit_cache = True


class lower(GenericFunction):
    type = String
    package = "odata"
    inherit_cache = True


class upper(GenericFunction):
    type = String
    package = "odata"
    inherit_cache = True


class ltrim(GenericFunction):
    type = String
    package = "odata"
    inherit_cache = True


class rtrim(GenericFunction):
    type = String
    package = "odata"
    inherit_cache = True


class ceil(GenericFunction):
    type = Integer
    package = "odata"
    inherit_cache = True


class floor(GenericFunction):
    type = Integer
    package = "odata"
    inherit_cache = True


class round(GenericFunction):
    type = Integer
    package = "odata"
    inherit_cache = True



import operator
from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Type, Union
from uuid import UUID

import django
from django.db.models import (
    Case,
    Exists,
    F,
    Model,
    OuterRef,
    Q,
    Value,
    When,
    functions,
    lookups,
)
from django.db.models.expressions import Expression

try:
    # Django gis requires system level libraries, which not every user needs.
    from django.contrib.gis.db.models import functions as gis_functions
    from django.contrib.gis.geos import GEOSGeometry

    _gis_error = None
except Exception as e:
    gis_functions = None
    GEOSGeometry = None
    _gis_error = e

from odata_query import ast, exceptions as ex, typing, utils, visitor

from .django_q_ext import NotEqual
from .utils import reverse_relationship

DJANGO_LT_4 = django.VERSION[0] < 4

COMPARISON_FLIP = {
    lookups.Exact: lookups.Exact,
    NotEqual: NotEqual,
    lookups.LessThan: lookups.GreaterThan,
    lookups.LessThanOrEqual: lookups.GreaterThanOrEqual,
    lookups.GreaterThan: lookups.LessThan,
    lookups.GreaterThanOrEqual: lookups.LessThanOrEqual,
    ast.Eq: ast.Eq,
    ast.NotEq: ast.NotEq,
    ast.Lt: ast.Gt,
    ast.LtE: ast.GtE,
    ast.Gt: ast.Lt,
    ast.GtE: ast.LtE,
}


def requires_gis(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not gis_functions:
            raise ImportError(
                "Cannot use geography functions because GeoDjango failed to load."
            ) from _gis_error
        return func(*args, **kwargs)

    return wrapper


class AstToDjangoQVisitor(visitor.NodeVisitor):
    """
    :class:`NodeVisitor` that transforms an :term:`AST` into a Django Q
    filter object.

    Args:
        root_model: The root model of the query.
    """

    def __init__(self, root_model: Type[Model]):
        self.root_model = root_model
        self.queryset_annotations: Dict[str, Expression] = {}

        # Keep track of the depth of `visit` calls, so we know when we should
        # turn the Django expression into a final `Q` object.
        self._depth: int = 0

    def visit(self, node: ast._Node) -> Any:
        """:meta private:"""
        self._depth += 1
        res = super().visit(node)
        self._depth -= 1

        if self._depth == 0:
            res = self._ensure_q(res)

        return res

    def visit_Identifier(self, node: ast.Identifier) -> F:
        ":meta private:"
        return F(node.name)

    def visit_Attribute(self, node: ast.Attribute) -> F:
        ":meta private:"
        owner = self.visit(node.owner)
        full_id = owner.name + "__" + node.attr
        return F(full_id)

    def visit_Null(self, node: ast.Null) -> str:
        ":meta private:"
        raise NotImplementedError("Should not be reached")

    def visit_Integer(self, node: ast.Integer) -> Value:
        ":meta private:"
        return Value(node.py_val)

    def visit_Float(self, node: ast.Float) -> Value:
        ":meta private:"
        return Value(node.py_val)

    def visit_Boolean(self, node: ast.Boolean) -> Value:
        ":meta private:"
        return Value(node.py_val)

    def visit_String(self, node: ast.String) -> Value:
        ":meta private:"
        return Value(node.py_val)

    def visit_Date(self, node: ast.Date) -> Value:
        ":meta private:"
        try:
            return Value(node.py_val)
        except ValueError:
            raise ex.ValueException(node.val)

    def visit_DateTime(self, node: ast.DateTime) -> Value:
        ":meta private:"
        try:
            return Value(node.py_val)
        except ValueError:
            raise ex.ValueException(node.val)

    def visit_Time(self, node: ast.Time) -> Value:
        ":meta private:"
        try:
            return Value(node.py_val)
        except ValueError:
            raise ex.ValueException(node.val)

    def visit_Duration(self, node: ast.Duration) -> Value:
        ":meta private:"
        return Value(node.py_val)

    def visit_GUID(self, node: ast.GUID) -> Value:
        ":meta private:"
        return Value(node.py_val)

    def visit_List(self, node: ast.List) -> List:
        ":meta private:"
        return [self.visit(n) for n in node.val]

    def visit_Add(self, node: ast.Add) -> Callable[[Any, Any], Any]:
        ":meta private:"
        return operator.add

    def visit_Sub(self, node: ast.Sub) -> Callable[[Any, Any], Any]:
        ":meta private:"
        return operator.sub

    def visit_Mult(self, node: ast.Mult) -> Callable[[Any, Any], Any]:
        ":meta private:"
        return operator.mul

    def visit_Div(self, node: ast.Div) -> Callable[[Any, Any], Any]:
        ":meta private:"
        return operator.truediv

    def visit_Mod(self, node: ast.Mod) -> Callable[[Any, Any], Any]:
        ":meta private:"
        return operator.mod

    def visit_BinOp(self, node: ast.BinOp) -> Any:
        ":meta private:"
        # Left or right can be an Identifier, in which case it needs to be
        # wrapped in F()
        left = self.visit(node.left)
        right = self.visit(node.right)
        op = self.visit(node.op)

        return op(left, right)

    def visit_Eq(self, node: ast.Eq) -> Type[lookups.Lookup]:
        ":meta private:"
        return lookups.Exact

    def visit_NotEq(self, node: ast.NotEq) -> Type[lookups.Lookup]:
        ":meta private:"
        return NotEqual

    def visit_Lt(self, node: ast.Lt) -> Type[lookups.Lookup]:
        ":meta private:"
        return lookups.LessThan

    def visit_LtE(self, node: ast.LtE) -> Type[lookups.Lookup]:
        ":meta private:"
        return lookups.LessThanOrEqual

    def visit_Gt(self, node: ast.Gt) -> Type[lookups.Lookup]:
        ":meta private:"
        return lookups.GreaterThan

    def visit_GtE(self, node: ast.GtE) -> Type[lookups.Lookup]:
        ":meta private:"
        return lookups.GreaterThanOrEqual

    def visit_In(self, node: ast.In) -> Type[lookups.Lookup]:
        ":meta private:"
        return lookups.In

    def visit_Compare(self, node: ast.Compare) -> lookups.Lookup:
        ":meta private:"
        lhs = self.visit(node.left)

        # Special case: comparison to NULL => isnull=True/False
        # Should not be wrapped with Value(True/False)
        # See: https://github.com/django/django/blob/0aacbdcf27b258387643b033352e99e6103abda8/django/db/models/lookups.py#L515
        if isinstance(node.right, ast.Null):
            if isinstance(node.comparator, ast.Eq):
                return lookups.IsNull(lhs, True)
            elif isinstance(node.comparator, ast.NotEq):
                return lookups.IsNull(lhs, False)
            else:
                raise ex.TypeException(node.comparator.__class__.__name__, "null")

        django_cls = self.visit(node.comparator)
        rhs = self.visit(node.right)

        return django_cls(lhs, rhs)

    def visit_And(self, node: ast.And) -> Callable[[Q, Q], Q]:
        ":meta private:"
        return operator.and_

    def visit_Or(self, node: ast.Or) -> Callable[[Q, Q], Q]:
        ":meta private:"
        return operator.or_

    def visit_BoolOp(self, node: ast.BoolOp) -> Q:
        ":meta private:"
        left = self.visit(node.left)
        right = self.visit(node.right)

        if isinstance(left, (F, Value)):
            raise ex.TypeException(node.op.__class__.__name__, left)
        if isinstance(right, (F, Value)):
            raise ex.TypeException(node.op.__class__.__name__, right)

        if DJANGO_LT_4:
            left = self._ensure_q(left)
            right = self._ensure_q(right)

        op = self.visit(node.op)

        return op(left, right)

    def visit_Not(self, node: ast.Not) -> Callable[[Q], Q]:
        ":meta private:"
        return operator.invert

    def visit_UnaryOp(self, node: ast.UnaryOp) -> str:
        ":meta private:"
        mod = self.visit(node.op)
        val = self.visit(node.operand)

        # Can only apply `~` to Q objects:
        val = self._ensure_q(val)

        try:
            return mod(val)
        except TypeError:
            raise ex.TypeException(node.op.__class__.__name__, val)

    def visit_Call(self, node: ast.Call) -> Union[Expression, Q]:
        ":meta private:"

        func_name = node.func.full_name().replace(".", "__")

        try:
            q_gen = getattr(self, "djangofunc_" + func_name.lower())
        except AttributeError:
            raise ex.UnsupportedFunctionException(func_name)

        args = []
        kwargs = {}
        for arg in node.args:
            if isinstance(arg, ast.NamedParam):
                kwargs[arg.name.name] = arg.param
            else:
                args.append(arg)

        res = q_gen(*args, **kwargs)
        return res

    def visit_CollectionLambda(self, node: ast.CollectionLambda) -> Q:
        ":meta private:"
        # NOTE: The initial implementation translated to SQL's ANY/ALL keywords,
        # but those behave very differently from OData's any/all keywords!

        owner_path = self._attempt_keywordify(node.owner)
        if not owner_path:
            raise ex.TypeException("lambda_expression", str(node.owner))

        path_to_outerref, related_model = reverse_relationship(
            owner_path, self.root_model
        )
        subquery = related_model.objects.filter(Q(**{path_to_outerref: OuterRef("pk")}))
        # .values(related_field.remote_field.name)

        if node.lambda_:
            # For the lambda, we want to strip the identifier off, because
            # we will execute this as a subquery in the wanted model's context.
            subq_ast = utils.expression_relative_to_identifier(
                node.lambda_.identifier, node.lambda_.expression
            )
            subq_transformer = self.__class__(related_model)
            subquery_filter = subq_transformer.visit(subq_ast)
        else:
            subquery_filter = None

        if isinstance(node.operator, ast.Any):
            # If ANY item should match in the subquery, we can use EXISTS():
            if subquery_filter:
                subquery = subquery.filter(subquery_filter)
            return Exists(subquery)

        elif isinstance(node.operator, ast.All):
            # If ALL items in the collection must match, we invert the condition and use NOT EXISTS():
            if subquery_filter:
                subquery = subquery.filter(~subquery_filter)
            return Exists(subquery, negated=True)

        else:
            raise NotImplementedError()

    @requires_gis
    def djangofunc_geo__intersects(
        self, field: ast.Identifier, geography: ast.Geography
    ):
        return Q(**{field.name + "__" + "intersects": GEOSGeometry(geography.wkt())})

    @requires_gis
    def djangofunc_geo__distance(self, field: ast.Identifier, geography: ast.Geography):
        return gis_functions.Distance(field.name, GEOSGeometry(geography.wkt()))

    @requires_gis
    def djangofunc_geo__length(self, field: ast.Identifier):
        return gis_functions.Length(field.name)

    def djangofunc_contains(
        self, field: ast._Node, substr: ast._Node
    ) -> lookups.Contains:
        ":meta private:"
        return self._substr_function(field, substr, lookups.Contains)

    def djangofunc_startswith(
        self, field: ast._Node, substr: ast._Node
    ) -> lookups.StartsWith:
        ":meta private:"
        return self._substr_function(field, substr, lookups.StartsWith)

    def djangofunc_endswith(
        self, field: ast._Node, substr: ast._Node
    ) -> lookups.EndsWith:
        ":meta private:"
        return self._substr_function(field, substr, lookups.EndsWith)

    def djangofunc_length(self, arg: ast._Node) -> functions.Length:
        ":meta private:"
        return functions.Length(self.visit(arg))

    def djangofunc_concat(self, *args: ast._Node) -> functions.Concat:
        ":meta private:"
        return functions.Concat(*[self.visit(arg) for arg in args])

    def djangofunc_indexof(
        self, first: ast._Node, second: ast._Node
    ) -> functions.StrIndex:
        ":meta private:"
        # Subtract 1 because OData is 0-indexed while SQL is 1-indexed
        return functions.StrIndex(self.visit(first), self.visit(second)) - 1

    def djangofunc_substring(
        self, fullstr: ast._Node, index: ast._Node, nchars: Optional[ast._Node] = None
    ) -> functions.Substr:
        ":meta private:"
        # Add 1 because OData is 0-indexed while SQL is 1-indexed
        return functions.Substr(
            self.visit(fullstr),
            self.visit(index) + 1,
            self.visit(nchars) if nchars else None,
        )

    def djangofunc_matchespattern(
        self, field: ast._Node, pattern: ast._Node
    ) -> lookups.Regex:
        ":meta private:"
        return lookups.Regex(self.visit(field), self.visit(pattern))

    def djangofunc_tolower(self, field: ast._Node) -> functions.Lower:
        ":meta private:"
        return functions.Lower(self.visit(field))

    def djangofunc_toupper(self, field: ast._Node) -> functions.Upper:
        ":meta private:"
        return functions.Upper(self.visit(field))

    def djangofunc_trim(self, field: ast._Node) -> functions.Trim:
        ":meta private:"
        return functions.Trim(self.visit(field))

    def djangofunc_date(self, field: ast._Node) -> functions.TruncDate:
        ":meta private:"
        return functions.TruncDate(self.visit(field))

    def djangofunc_day(self, field: ast._Node) -> functions.ExtractDay:
        ":meta private:"
        return functions.ExtractDay(self.visit(field))

    def djangofunc_hour(self, field: ast._Node) -> functions.ExtractHour:
        ":meta private:"
        return functions.ExtractHour(self.visit(field))

    def djangofunc_minute(self, field: ast._Node) -> functions.ExtractMinute:
        ":meta private:"
        return functions.ExtractMinute(self.visit(field))

    def djangofunc_month(self, field: ast._Node) -> functions.ExtractMonth:
        ":meta private:"
        return functions.ExtractMonth(self.visit(field))

    def djangofunc_now(self) -> functions.Now:
        ":meta private:"
        return functions.Now()

    def djangofunc_second(self, field: ast._Node) -> functions.ExtractSecond:
        ":meta private:"
        return functions.ExtractSecond(self.visit(field))

    def djangofunc_time(self, field: ast._Node) -> functions.TruncTime:
        ":meta private:"
        return functions.TruncTime(self.visit(field))

    def djangofunc_year(self, field: ast._Node) -> functions.ExtractYear:
        ":meta private:"
        return functions.ExtractYear(self.visit(field))

    def djangofunc_ceiling(self, field: ast._Node) -> functions.Ceil:
        ":meta private:"
        return functions.Ceil(self.visit(field))

    def djangofunc_floor(self, field: ast._Node) -> functions.Floor:
        ":meta private:"
        return functions.Floor(self.visit(field))

    def djangofunc_round(self, field: ast._Node) -> functions.Round:
        ":meta private:"
        return functions.Round(self.visit(field))

    def _substr_function(
        self, field: ast._Node, substr: ast._Node, django_func: Type[Expression]
    ) -> Expression:
        ":meta private:"
        typing.typecheck(field, (ast.Identifier, ast.String), "field")
        typing.typecheck(substr, ast.String, "substring")

        return django_func(self.visit(field), self.visit(substr))

    def _fix_uuid(self, node: Any) -> Any:
        # Workaround for Django <4 'Value is not a valid UUID':
        if isinstance(node, Value) and isinstance(node.value, UUID):
            return node.value

        if isinstance(node, list):
            return [self._fix_uuid(i) for i in node]

        return node

    def _ensure_q(self, node: Any) -> Q:
        """
        Turn a given Django `Lookup`, `Expression` or `Function` into a `Q` object.
        This is mainly necessary for Django <4 where the nodes cannot be directly
        used in `Q` objects and have to be expressed as `Q(keyword=value)`.
        """
        if isinstance(node, (Q, Exists)):
            return node

        if not DJANGO_LT_4:
            return Q(node)

        # Need an identifier on any side to make a Django Q:
        keyword = self._attempt_keywordify(node.lhs)
        if not keyword:
            # No keyword on the left, try the right:
            keyword = self._attempt_keywordify(node.rhs)
            if keyword:
                node = self._flip_comparison(node)
            else:
                # No keywords at all, cannot continue:
                raise ex.TypeException(node.__class__.__name__, str(node.lhs))

        if isinstance(node, lookups.Lookup):
            keyword += "__" + node.lookup_name

        node.rhs = self._fix_uuid(node.rhs)
        query = Q(**{keyword: node.rhs})
        return query

    def _attempt_keywordify(self, node: Any) -> Optional[str]:
        """
        Try to turn ``node`` into a keyword argument that can be used in a Django
        ``Q`` object. E.g. a ``contains(name, 'something')`` node should resolve
        to the keyword ``name__contains``.

        :meta private:
        """
        # A literal can not be expressed as a keyword.
        if isinstance(node, (ast._Literal, Value)):
            return None

        # If an AST Node was passed, visit it to get something Django related:
        if isinstance(node, ast._Node):
            res = self.visit(node)
        else:
            res = node

        # If `res` is already wrapped in a `Q` object, we need to unwrap it first.
        # This is the case with expressions that are filterable by themselves,
        # such as `contains(a, b)`.
        if isinstance(res, Q) and isinstance(res.children[0], Expression):
            res = res.children[0]

        # An `F` expression is a field or expression known to Django, and can
        # be used as a keyword.
        if isinstance(res, F):
            return res.name

        # Field lookups are also easily expressed as keywords.
        if (
            hasattr(res, "lookup_name")
            and hasattr(res, "lhs")
            and hasattr(res.lhs, "name")
            # Expressions with a `rhs` have parameters and should be handled
            # as function calls:
            and not getattr(res, "rhs", False)
        ):
            return res.lhs.name + "__" + res.lookup_name

        # Lookups with parameters need to be wrapped in a `CASE WHEN` expression
        if isinstance(res, lookups.Lookup):
            identity = self._gen_annotation_name(res)
            if DJANGO_LT_4:
                res = self._ensure_q(res)
            res = Case(When(res, then=True), default=False)
            self.queryset_annotations[identity] = res
            return identity

        # For more complicated expressions, we can add them to the query as a
        # QuerySet annotation. This annotation is then valid as a keyword:
        if isinstance(res, Expression):
            identity = self._gen_annotation_name(res)
            self.queryset_annotations[identity] = res
            return identity

        return None

    def _gen_annotation_name(self, expr: Expression) -> str:
        ":meta private:"
        if hasattr(expr, "name"):
            return expr.name
        elif hasattr(expr, "value"):
            return str(expr.value)

        func_name = expr.__class__.__name__

        try:
            args = expr.get_source_expressions()
        except AttributeError:
            args = []

        args_str = [self._gen_annotation_name(a) for a in args]

        return (
            "_".join([func_name] + args_str)
            .replace(" ", "_")
            .replace(",", "")
            .replace(":", "_")
            .lower()
        )

    @staticmethod
    def _flip_comparison(comp: lookups.Lookup) -> lookups.Lookup:
        """
        Flip a comparison left-to-right. E.g.: (4 > version_id) becomes (version_id < 4)

        :meta private:
        """
        new_op = COMPARISON_FLIP[type(comp)]
        new_left = comp.rhs
        new_right = comp.lhs

        return new_op(new_left, new_right)



import re

from odata_query import ast, exceptions, typing

from .base import AstToSqlVisitor

UNSAFE_CHARS = re.compile(r"[^a-zA-Z0-9_]")


def clean_athena_identifier(identifier: str) -> str:
    """
    Cleans an Athena identifier so it passes the following validation rules:

    - Table names and table column names in Athena must be lowercase
    - Athena table, view, database, and column names allow only underscore special characters
    - Names should be quoted or backticked when starting with a number or underscore

    Source: https://docs.aws.amazon.com/athena/latest/ug/tables-databases-columns-names.html
    """
    id_new = identifier.lower()
    id_new = UNSAFE_CHARS.sub("_", id_new)
    return id_new


class AstToAthenaSqlVisitor(AstToSqlVisitor):
    """
    :class:`NodeVisitor` that transforms an :term:`AST` into an Athena SQL
    ``WHERE`` clause.

    Args:
        table_alias: Optional alias for the root table.
    """

    def visit_Identifier(self, node: ast.Identifier) -> str:
        ":meta private:"
        # Double quotes for column names acc SQL Standard
        sql_id = f'"{clean_athena_identifier(node.name)}"'

        if self.table_alias:
            sql_id = f'"{self.table_alias}".' + sql_id

        return sql_id

    def visit_DateTime(self, node: ast.DateTime) -> str:
        ":meta private:"
        return f"FROM_ISO8601_TIMESTAMP('{node.val}')"

    def sqlfunc_length(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        inferred_type = typing.infer_type(arg)

        # If the input is a string or default, assume str-length:
        if inferred_type is ast.String or inferred_type is None:
            return f"LENGTH({arg_sql})"

        # If the input is a list, assume list-length:
        if inferred_type is ast.List:
            return f"CARDINALITY({arg_sql})"

        raise exceptions.ArgumentTypeException("length")

    def sqlfunc_substring(self, *args: ast._Node) -> str:
        ":meta private:"
        args_sql = [self.visit(arg) for arg in args]
        inferred_type = typing.infer_type(args[0])

        # If the first input is a string or default, assume str-substr:
        if inferred_type is ast.String or inferred_type is None:
            if len(args) == 2:
                return f"SUBSTR({args_sql[0]}, {args_sql[1]} + 1)"
            if len(args) == 3:
                return f"SUBSTR({args_sql[0]}, {args_sql[1]} + 1, {args_sql[2]})"

        # If the first input is a list, assume list-substr:
        if inferred_type is ast.List:
            if len(args) == 2:
                return f"SLICE({args_sql[0]}, {args_sql[1]})"
            if len(args) == 3:
                return f"SLICE({args_sql[0]}, {args_sql[1]}, {args_sql[2]})"

        raise exceptions.ArgumentTypeException("substring")

    def sqlfunc_round(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"ROUND({arg_sql})"

    def sqlfunc_floor(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"FLOOR({arg_sql})"

    def sqlfunc_ceiling(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"CEILING({arg_sql})"

    def sqlfunc_hassubset(self, *args: ast._Node) -> str:
        ":meta private:"
        args_sql = [self.visit(arg) for arg in args]
        return f"CARDINALITY(ARRAY_INTERSECT({args_sql[0]}, {args_sql[1]})) = CARDINALITY({args_sql[1]})"



import sys
from typing import Type

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol

from odata_query import ast, visitor


class LiteralValNode(Protocol):
    """:meta private:"""

    val: str


PRECEDENCE = {
    ast.Attribute: 10,
    ast.Call: 10,
    ast.Not: 9,
    ast.USub: 9,
    ast.Mult: 8,
    ast.Div: 8,
    ast.Mod: 8,
    ast.Add: 7,
    ast.Sub: 7,
    ast.Gt: 6,
    ast.GtE: 6,
    ast.Lt: 6,
    ast.LtE: 6,
    ast.Eq: 5,
    ast.NotEq: 5,
    ast.And: 4,
    ast.Or: 3,
}


class AstToODataVisitor(visitor.NodeVisitor):
    """
    :class:`NodeVisitor` that transforms an :term:`AST` back into an OData
    query string.
    """

    def visit_Identifier(self, node: ast.Identifier) -> str:
        """:meta private:"""
        if node.namespace:
            prefix = ".".join(node.namespace)
            return prefix + "." + node.name
        return node.name

    def visit_Attribute(self, node: ast.Attribute) -> str:
        """:meta private:"""
        return self.visit(node.owner) + "/" + node.attr

    def visit_Null(self, node: ast.Null) -> str:
        """:meta private:"""
        return "null"

    def visit_String(self, node: ast.String) -> str:
        """:meta private:"""
        return "'" + node.val + "'"

    def visit_Duration(self, node: ast.Duration) -> str:
        """:meta private:"""
        return "duration'" + node.val + "'"

    def _visit_Literal(self, node: LiteralValNode) -> str:
        """:meta private:"""
        return node.val

    visit_Integer = _visit_Literal
    visit_Float = _visit_Literal
    visit_Boolean = _visit_Literal
    visit_Date = _visit_Literal
    visit_Time = _visit_Literal
    visit_DateTime = _visit_Literal
    visit_GUID = _visit_Literal

    def visit_List(self, node: ast.List) -> str:
        """:meta private:"""
        return "(" + ", ".join(self.visit(v) for v in node.val) + ")"

    def visit_Add(self, node: ast.Add) -> str:
        """:meta private:"""
        return "add"

    def visit_Sub(self, node: ast.Sub) -> str:
        """:meta private:"""
        return "sub"

    def visit_Mult(self, node: ast.Mult) -> str:
        """:meta private:"""
        return "mul"

    def visit_Div(self, node: ast.Div) -> str:
        """:meta private:"""
        return "div"

    def visit_Mod(self, node: ast.Mod) -> str:
        """:meta private:"""
        return "mod"

    def visit_BinOp(self, node: ast.BinOp) -> str:
        """:meta private:"""
        left = self._visit_and_paren_if_precedence_lower(node.left, type(node.op))
        right = self._visit_and_paren_if_precedence_lower(node.right, type(node.op))
        return left + " " + self.visit(node.op) + " " + right

    def visit_Eq(self, node: ast.Eq) -> str:
        """:meta private:"""
        return "eq"

    def visit_NotEq(self, node: ast.NotEq) -> str:
        """:meta private:"""
        return "ne"

    def visit_Lt(self, node: ast.Lt) -> str:
        """:meta private:"""
        return "lt"

    def visit_LtE(self, node: ast.LtE) -> str:
        """:meta private:"""
        return "le"

    def visit_Gt(self, node: ast.Gt) -> str:
        """:meta private:"""
        return "gt"

    def visit_GtE(self, node: ast.GtE) -> str:
        """:meta private:"""
        return "ge"

    def visit_In(self, node: ast.In) -> str:
        """:meta private:"""
        return "in"

    def visit_Compare(self, node: ast.Compare) -> str:
        """:meta private:"""
        left = self._visit_and_paren_if_precedence_lower(
            node.left, type(node.comparator)
        )
        right = self._visit_and_paren_if_precedence_lower(
            node.right, type(node.comparator)
        )
        return left + " " + self.visit(node.comparator) + " " + right

    def visit_And(self, node: ast.And) -> str:
        """:meta private:"""
        return "and"

    def visit_Or(self, node: ast.Or) -> str:
        """:meta private:"""
        return "or"

    def visit_BoolOp(self, node: ast.BoolOp) -> str:
        """:meta private:"""
        left = self._visit_and_paren_if_precedence_lower(node.left, type(node.op))
        right = self._visit_and_paren_if_precedence_lower(node.right, type(node.op))
        return left + " " + self.visit(node.op) + " " + right

    def visit_Not(self, node: ast.Not) -> str:
        """:meta private:"""
        return "not"

    def visit_USub(self, node: ast.USub) -> str:
        """:meta private:"""
        return "-"

    def visit_UnaryOp(self, node: ast.UnaryOp) -> str:
        """:meta private:"""
        operand = self._visit_and_paren_if_precedence_lower(node.operand, type(node.op))
        return self.visit(node.op) + " " + operand

    def visit_Call(self, node: ast.Call) -> str:
        """:meta private:"""
        return (
            self.visit(node.func)
            + "("
            + ", ".join(self.visit(n) for n in node.args)
            + ")"
        )

    def visit_Any(self, node: ast.Any) -> str:
        """:meta private:"""
        return "any"

    def visit_All(self, node: ast.All) -> str:
        """:meta private:"""
        return "all"

    def visit_Lambda(self, node: ast.Lambda) -> str:
        """:meta private:"""
        return self.visit(node.identifier) + ": " + self.visit(node.expression)

    def visit_CollectionLambda(self, node: ast.CollectionLambda) -> str:
        """:meta private:"""
        return (
            self.visit(node.owner)
            + "/"
            + self.visit(node.operator)
            + "("
            + (self.visit(node.lambda_) if node.lambda_ else "")
            + ")"
        )

    def _visit_and_paren_if_precedence_lower(
        self, node: ast._Node, precedence: Type[ast._Node]
    ) -> str:
        """
        Transform `node` by visiting it, then wrap the result in parentheses if
        the expressions precedence is lower than that of `precedence`.

        :meta private:
        """
        res = self.visit(node)

        if hasattr(node, "op"):
            node_op = type(node.op)  # type: ignore
        elif hasattr(node, "comparator"):
            node_op = type(node.comparator)  # type: ignore
        else:
            node_op = type(node)

        node_prec = PRECEDENCE.get(node_op, 100)
        check_prec = PRECEDENCE.get(precedence, 100)

        if node_prec < check_prec:
            res = "(" + res + ")"

        return res
